{
FatturaElettronicaExtractor - Nicola Perotto 2019

Usa le seguenti librerie:
  JvString.pas      parte di JEDI VCL http://jvcl.sourceforge.net

  OpenSSLUtils.pas  https://github.com/bambucode/tfacturaelectronica/tree/master/OpenSSL
                    "With a contribute of Univerity of Genoa (Italy) http://www.unige.it/"

  Serve il file libeay32.dll che fa parte di OpenSSL.

  Versione 1.1  2019.01.18
    + Ripulisce il nome degli allegati da caratteri strani
    + Gestisce codifiche base 64 con gli a capo
  Versione 1.2  2019.01.23
    + config file
    + percorso predefinito dove salvare i file
    + switch per spostare .p7m
    + switch per aprire .xml (shell)
    + switch per rinominare file aggiungendo dati fattura (il .p7m resta con il suo nome)
        nuovo nome := vecchio + ' ' + RagioneSociale + DataFattura + NumeroFattura
  Versione 1.3  2019.02.07
    + decodifica correttamente anche i .p7m Base64 encoded
  Versione 1.4  2019.05.23
    * ora estrae tutti gli allegati non solo il primo
  Versione 1.5  2019.06.16
    * non va pi  in crash se il file  sballato
    * modificato i codici di errore
    * non crea pi la cartella se non ci sono allegati
}

Program FEExtractor;

{$APPTYPE CONSOLE}

Uses
  SysUtils, Windows, JvStrings, OpenSSLUtils, XMLIntf, XMLDoc, ComObj, ActiveX, Classes, IniFiles, ShellAPI;

Const
  Versione = 'FEExtractor 1.5';

  ErrNone                 = 0;
  ErrPrintHelp            = 1;
  ErrRenameError          = 2;
  ErrFileNotExists        = 3;
  ErrNoDestinationFolder  = 4;
  ErrFileMove             = 5;
  ErrSignature            = 6;
  ErrInvalidExtension     = 7;
  ErrShellOpen            = 8;

Var
  cfgDestFolder         : string;       //dove spostare i file
  cfgSposta             : Boolean;      //sposto?
  cfgApre               : Boolean;      //apro?
  cfgRinomina           : Boolean;      //rinomino xml e relativa cartella?

  NomeVecchio, Destinazione, Estensione, NomeFile : string;
  SEInfo: TShellExecuteInfo;
  j : integer;  //loop


Procedure ReadConfig;
Var
  Ini : TIniFile;
begin
  try
    Ini := TIniFile.Create(ChangeFileExt(ParamStr(0), '.ini'));
    cfgDestFolder := Ini.ReadString('Opzioni', 'Destinazione', '');
    if (cfgDestFolder <> '') and (cfgDestFolder[Length(cfgDestFolder)] <> '\') then cfgDestFolder := cfgDestFolder + '\';
    cfgSposta   := Ini.ReadBool('Opzioni', 'Sposta', cfgSposta); //essendo lette dopo il controllo dei parametri
    cfgApre     := Ini.ReadBool('Opzioni', 'Apre', cfgApre);     //queste impostazioni, se ci sono nel file, hanno priorit 
    cfgRinomina := Ini.ReadBool('Opzioni', 'Rinomina', cfgRinomina);
  finally
    FreeAndNil(Ini);
  end;
end; //ReadConfig;

Function LanciaXML:Boolean;
begin
  Result := False;
  try
    FillChar(SEInfo, SizeOf(SEInfo), 0);
    SEInfo.cbSize := SizeOf(TShellExecuteInfo) ;
    with SEInfo do begin
      fMask := SEE_MASK_NOCLOSEPROCESS;
      Wnd := 0; //Application.Handle;
      lpFile := PChar(NomeFile);
      nShow := SW_HIDE; //SW_SHOWNORMAL;
    end;
    if not ShellExecuteEx(@SEInfo) then Exit;
    Result := True;
  except
  end;
end; //LanciaXML

Function ToglieACapo(Const Base:string):string;
Var
  i, j : integer;
begin
  i := Pos(#13, Base);
  j := Pos(#10, Base); //ho beccato un allegato con solo LF
  if (i = 0) and (j = 0) then begin
    Result := Base;
    Exit;
  end;
  SetLength(Result, Length(Base));
  j := 0;
  for i := 1 to Length(Base) do begin
    if Ord(Base[i]) > 32 then begin//molto brutale...
      Inc(j);
      Result[j] := Base[i];
    end;
  end;
  SetLength(Result, j);
end; //ToglieACapo

Function NormalizzaNomeFile(Const NomeStrano:string):string;
Var
  i : integer;
begin
  Result := NomeStrano;
  for i := 1 to Length(NomeStrano) do begin
    if Pos(NomeStrano[i], '<>:"\|?*') > 0 then
      Result[i] := '_'
    else if NomeStrano[i] = '/' then Result[i] := '-'; //mi piace di pi!
  end;
end; //NormalizzaNomeFile

Procedure EstraiAllegati;
Var
  Cartella : string;
  XML : IXMLDOCUMENT;
  Node1, Node2, Node3, Node4, Node5 : IXMLNODE;
  i, n : integer;
  Testo, Normale, NomeAllegato, FormatoAllegato : string;
  f : TFileStream;
  RagioneSociale, NumeroFattura, DataFattura : string;
begin
  try
    CoInitializeEx(nil, COINIT_APARTMENTTHREADED); //Inizializza il sottosistema COM per poter usare le 'Interface'
    XML := NewXMLDocument;
    XML.Encoding := 'UTF-8';
    XML.NodeIndentStr := '  '; //space *2
    XML.Options := [doNodeAutoIndent]; // looks better in Editor ;)
    XML.LoadFromFile(NomeFile);

//2              <1.N>  <FatturaElettronicaBody>
//2.5            <0.N>  <Allegati>
//2.5.1          <1.1>  <NomeAttachment>  xs:normalizedString 1  60
//2.5.2          <0.1>  <AlgoritmoCompressione>  xs:string 1  10
//2.5.3          <0.1>  <FormatoAttachment>  xs:string 1  10
//2.5.4          <0.1>  <DescrizioneAttachment>  xs:normalizedString 1  100
//2.5.5          <1.1>  <Attachment>  xs:base64Binary valore vincolato alla dimensione max prevista per la fattura elettronica

    if CompareText(XML.DocumentElement.LocalName, 'FatturaElettronica') <> 0 then begin
      writeln('Non  una Fattura Elettronica.');
      Exit;
    end;

    if cfgRinomina then begin
      Node1 := XML.DocumentElement.ChildNodes.FindNode('FatturaElettronicaHeader', ''); //1              <1.1>  <FatturaElettronicaHeader>
      if not Assigned(Node1) then Exit;
      Node2 := Node1.ChildNodes.FindNode('CedentePrestatore', ''); //1.2            <1.1>  <CedentePrestatore>
      if not Assigned(Node2) then Exit;
      Node3 := Node2.ChildNodes.FindNode('DatiAnagrafici', ''); //1.2.1          <1.1>  <DatiAnagrafici>
      if not Assigned(Node3) then Exit;
      Node4 := Node3.ChildNodes.FindNode('Anagrafica', ''); //1.2.1.3        <1.1>  <Anagrafica>
      if not Assigned(Node4) then Exit;
      Node5 := Node4.ChildNodes.FindNode('Denominazione', ''); //1.2.1.3.1      <0.1>  <Denominazione>  xs:normalizedString 1  80
      if not Assigned(Node5) then begin
        Node5 := Node4.ChildNodes.FindNode('Nome', ''); //1.2.1.3.2      <0.1>  <Nome>  xs:normalizedString 1  60
        if not Assigned(Node5) then Exit;
        RagioneSociale := Node5.NodeValue;
        Node5 := Node4.ChildNodes.FindNode('Cognome', ''); //1.2.1.3.3      <0.1>  <Cognome>  xs:normalizedString 1  60
        if not Assigned(Node5) then Exit;
        RagioneSociale := Node5.NodeValue + ' ' + RagioneSociale;
      end else begin
        RagioneSociale := Node5.NodeValue;
      end;
    end;

    Node1 := XML.DocumentElement.ChildNodes.FindNode('FatturaElettronicaBody', ''); //2              <1.N>  <FatturaElettronicaBody>   
    if not Assigned(Node1) then Exit;

    if cfgRinomina then begin
      Node2 := Node1.ChildNodes.FindNode('DatiGenerali', ''); //2.1            <1.1>  <DatiGenerali>
      if not Assigned(Node2) then Exit;
      Node3 := Node2.ChildNodes.FindNode('DatiGeneraliDocumento', ''); //2.1.1          <1.1>  <DatiGeneraliDocumento>   
      if not Assigned(Node3) then Exit;
      Node4 := Node3.ChildNodes.FindNode('Data', ''); //2.1.1.3        <1.1>  <Data>  xs:date 10
      if not Assigned(Node4) then Exit;
      DataFattura := Node4.NodeValue;
      Node4 := Node3.ChildNodes.FindNode('Numero', ''); //2.1.1.4        <1.1>  <Numero>  xs:normalizedString 1  20
      if not Assigned(Node4) then Exit;
      NumeroFattura := Node4.NodeValue;
    end;

    if cfgRinomina then begin
      NomeVecchio := NomeFile; //l'XML  aperto, lo rinomino per ultimo
      RagioneSociale := NormalizzaNomeFile(RagioneSociale);
      NumeroFattura := NormalizzaNomeFile(NumeroFattura); //QUESTO serve!
      DataFattura := NormalizzaNomeFile(DataFattura); //questo mooooolto probabilmente no
      NomeFile := ChangeFileExt(NomeFile, '') + ' ' + RagioneSociale + ' ' + DataFattura + ' ' + NumeroFattura + '.xml';
    end;

    //Se ci sono 1 o pi allegati, li salvo nella cartella con lo stesso nome del file XML
    Cartella := ChangeFileExt(NomeFile, '');
    Cartella := Cartella + '\';
    n := 1;
    Node2 := Node1.ChildNodes.FindNode('Allegati', '');
    while Assigned(Node2) do begin
      if not DirectoryExists(Cartella) then CreateDir(Cartella);
      NomeAllegato := '';
      FormatoAllegato := '';
      i := 0;
      while (i <= Node2.ChildNodes.Count - 1) do begin
        Node3 := Node2.ChildNodes.Get(i);
        if CompareText(Node3.LocalName, 'NomeAttachment') = 0 then
          NomeAllegato := Node3.NodeValue;
        if CompareText(Node3.LocalName, 'AlgoritmoCompressione') = 0 then ; //ignora
        if CompareText(Node3.LocalName, 'FormatoAttachment') = 0 then begin
          FormatoAllegato := Node3.NodeValue;
        end;
        if CompareText(Node3.LocalName, 'DescrizioneAttachment') = 0 then ; //ignora
        if CompareText(Node3.LocalName, 'Attachment') = 0 then begin
          Testo := Node3.NodeValue;
          try
            Normale := B64Decode(ToglieACapo(Testo));
          except
            on E:Exception do begin
              writeln('L''allegato ' + IntToStr(n) + ' non  codificato Base64 correttamente.');
              Normale := '';
            end;
          end;
          if (NomeAllegato <> '') and (Normale <> '') then begin
            try
              NomeAllegato := NormalizzaNomeFile(NomeAllegato);
              FormatoAllegato := NormalizzaNomeFile(FormatoAllegato);
              if ExtractFileExt(NomeAllegato) = '' then //provo a dargli anche un formato
                NomeAllegato := NomeAllegato + '.' + FormatoAllegato;
              f := TFileStream.Create(Cartella + NomeAllegato, fmCreate);
              f.Write(Normale[1], Length(Normale));
            finally
              FreeAndNil(f);
            end;
          end;
          NomeAllegato := '';
        end; //if
        Inc(i);
      end; //for i
      Inc(n);
      Node2 := Node2.NextSibling;
    end; //while (ci sono allegati)
  finally
    XML.Active := False;
    if cfgRinomina and not MoveFileEx(PChar(NomeVecchio), PChar(NomeFile), MOVEFILE_COPY_ALLOWED + MOVEFILE_REPLACE_EXISTING + MOVEFILE_WRITE_THROUGH) then begin
      writeln('Errore durante il cambio di nome: ' + IntToStr(GetLastError));
      Halt(ErrRenameError);
    end;
//    CoUninitialize; //se lo chiamo poi si incazza!
  end;
end; //EstraiAllegati

Const
  Valid64 = ['A'..'Z', 'a'..'z', '0'..'9', '+', '/', '='];
//  B64Table: AnsiString = 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/=';

Function IsValidP7M(Var NomeFile:string):boolean;
Var
  t : TStringList;
  f : TFileStream;
  i : integer;
  s : string;
begin
  Result := False;
  t := TStringList.Create;
  f := Nil;
  try
    t.LoadFromFile(NomeFile);
//UTF-8 signature: $EF $BB $BF
    s := copy(t[0], 1, 12);
    if Length(s) < 3 then Exit;
    if ((s[1] = #$EF) and (s[2] = #$BB) and (s[3] = #$BF)) then begin
      //File UTF-8, forse senza firma!
    end else begin
      for i := 1 to 4 do begin
        if not (s[i] in Valid64) then begin
          Result := True;
          Exit; //non sembra Base64
        end;
      end;
      //Sembra Base64
      //Verifico che non abbia CR e/o LF
      if t.Count > 1 then
        s := ToglieACapo(t.Text)
      else
        s := t.Text;
      s := B64Decode(ToglieACapo(t.Text));
      NomeFile := ChangeFileExt(ParamStr(0), '.tmp.p7m'); //temporaneo!
      f := TFileStream.Create(NomeFile, fmCreate);
      f.Write(s[1], Length(s));
    end;
    Result := True;
  finally
    FreeAndNil(t);
    FreeAndNil(f);
  end;
end; //IsValidP7M

Function TogliFirma(Nomep7m, NomeXml:string):Boolean;
Var
  Reader : TPKCS7;
begin
  Result := False;
  //controlla che di tipo di file si tratti:
  //- normale
  //- base64
  if not IsValidP7M(Nomep7m) then
    Exit;
  try
    Reader := TPKCS7.Create;
    Reader.Open(Nomep7m);
    Reader.SaveContent(NomeXml);
    Reader.Free;
    Result := True;
  except
  end;
end; //TogliFirma

begin
  cfgDestFolder         := '';
  cfgSposta             := False;
  cfgApre               := False;
  cfgRinomina           := False;

  if ParamCount = 0 then begin
    writeln(Versione);
    writeln('Estrae gli allegati da una fattura elettronica, eventualmente rimuovendo la firma.');
    writeln('Attenzione: sovrascrive SENZA AVVISARE!');
    writeln('Uso: FEExtractor [/a] [/s] [r] nomefile');
    writeln('/a : apre il file .xml usando il programma predefinito');
    writeln('/s : sposta il file nella cartella impostata nel file .ini');
    writeln('/r : rinominca il xile xml e la cartella aggiungendo RagioneSociale, DataFattura e NumeroFattura');
    Halt(errPrintHelp);
  end;

  //controlla i parametri
  for j := 1 to ParamCount do begin
    if (CompareText('/a', ParamStr(j)) = 0) then
      cfgApre := True
    else if (CompareText('/s', ParamStr(j)) = 0) then
      cfgSposta := True
    else if (CompareText('/r', ParamStr(j)) = 0) then
      cfgRinomina := True
    else if (NomeFile = '') and FileExists(ParamStr(j)) then
      NomeFile := ParamStr(j);
  end;

  if not FileExists(NomeFile) then begin
    writeln('Errore: il file <' + NomeFile + '> non esiste.');
    Halt(ErrFileNotExists);
  end;
  Estensione := ExtractFileExt(NomeFile);
  ReadConfig;
  if cfgSposta and ((cfgDestFolder = '') or not DirectoryExists(cfgDestFolder)) then begin
    writeln('Errore: manca la cartella di destinazione.');
    Halt(ErrNoDestinationFolder);
  end;

  if cfgSposta and (CompareText(cfgDestFolder, ExtractFilePath(NomeFile)) = 0) then
    cfgSposta := False; // gi nella cartella giusta

  if cfgSposta then begin
    NomeVecchio := NomeFile;
    NomeFile := cfgDestFolder + ExtractFileName(NomeVecchio);
    if not MoveFileEx(PChar(NomeVecchio), PChar(NomeFile), MOVEFILE_COPY_ALLOWED + MOVEFILE_REPLACE_EXISTING + MOVEFILE_WRITE_THROUGH) then begin
      writeln('Errore durante lo spostamento: ' + IntToStr(GetLastError));
      Halt(ErrFileMove);
    end;
  end;

  if (CompareText(Estensione, '.xml') = 0) then begin
    EstraiAllegati;
  end else if (CompareText(Estensione, '.p7m') = 0) then begin
    Destinazione := ChangeFileExt(NomeFile, '');
    AppStartup;
    if not TogliFirma(NomeFile, Destinazione) then Halt(ErrSignature);
    NomeFile := Destinazione;
    EstraiAllegati;
  end else begin
    writeln('Errore: il file deve essere .xml oppure .xml.p7m');
    Halt(ErrInvalidExtension);
  end;
  if cfgApre then begin
    if not LanciaXML then begin
      writeln('Errore shell: ' + IntToStr(GetLastError));
      Halt(ErrShellOpen);
    end;
  end;
  Halt(ErrNone);
end.

Esistono parecchie casistiche.
(1) file di testo XML e basta [vabb]
(2) file di testo XML, con apposta firma XAdES
(3) file di testo XML, con apposta firma CAdES
(4) file di testo XML, con apposta firma CAdES codificata mime
(5) file XML codificato mime
(6) file XML codificato mime, con firma CAdES
(7) file XML codificato mime, con firma CAdES, codificata mime
etc...

